import re
import json
from datetime import datetime
import collections

from . import Serializer


class CellarDoorJSONEncoder(json.JSONEncoder):
	
	def default(self, obj):
		if isinstance(obj, collections.Iterable):
			return list(obj)
		
		if isinstance(obj, datetime):
			return obj.isoformat()
		
		return super(CellarDoorJSONEncoder, self).default(obj)
		
		
def as_date(obj):
	if '_date' in obj:
		return datetime(*map(int, re.split('[^\d]', obj['_date'])[:-1]))
	else:
		return obj
		


class JSONSerializer(Serializer):
	
	mimetype = 'application/json'
	
	def serialize(self, obj):
		return json.dumps(obj, cls=CellarDoorJSONEncoder)
		
		
	def unserialize(self, stream):
		return json.load(stream, object_hook=as_date)
		
		
	def unserialize_string(self, data):
		return json.loads(data, object_hook=as_date)


import types
import collections
from copy import deepcopy
import inspect
from ..model import ListOf, InverseLink
from ..events import EventManager
from .. import errors
from .methods import *


__all__ = [
	'Interface'
]


class RuleSet(object):
	
	def __init__(self, method_authorization):
		self.enabled_methods = set()
		self.item_rules = {}
		self.non_item_rules = {}
		
		if method_authorization:
			for k,v in method_authorization.items():
				if not isinstance(k, tuple):
					k = (k,)
				for method in k:
					self.enabled_methods.add(method)
					if v is None:
						continue
					if v.uses('item'):
						rules = self.item_rules
					else:
						rules = self.non_item_rules
					if method not in rules:
						rules[method] = []
					rules[method].append(v)
				
				
	def enforce_item_rules(self, method, item, context):
		rules = self.item_rules.get(method)
		if rules:
			if isinstance(item, (collections.Sequence, types.GeneratorType)):
				for i in item:
					self.enforce_rules(rules, i, context)
			else:
				self.enforce_rules(rules, item, context)
		
		
	def enforce_non_item_rules(self, method, context):
		rules = self.non_item_rules.get(method)
		if rules:
			self.enforce_rules(rules, None, context)
		
		
	def enforce_rules(self, rules, item, context):
		context = context if context else {}
		context['item'] = item
		no_identity = 'identity' not in context
		for rule in rules:
			if no_identity and rule.uses('identity'):
				raise errors.NotAuthenticatedError()
			if not rule(context):
				raise errors.NotAuthorizedError()
				
				
				
class OptionsFactory(object):
	
	
	def __init__(self, storage=None,
					   hidden_fields=(), 
					   hidden_field_authorization=None,
					   enabled_filters=(), 
					   enabled_sort=(), 
					   default_sort=(), 
					   default_limit=0, 
					   max_limit=0):
		self.storage = storage
		self.hidden_fields = set(hidden_fields)
		self.hidden_field_authorization = hidden_field_authorization
		self.enabled_filters = set(enabled_filters)
		self.enabled_filters_no_hidden = self.enabled_filters.difference(self.hidden_fields)
		self.enabled_sort = set(enabled_sort)
		self.enabled_sort_no_hidden = self.enabled_sort.difference(self.hidden_fields)
		self.default_sort = default_sort
		self.default_limit = default_limit
		self.max_limit = max_limit
		
		self.enabled_filters.update(('_id', '_type'))
		self.enabled_filters_no_hidden.update(('_id', '_type'))
		
		
	def create(self, options_dict, list=False):
		copied_options_dict = deepcopy(options_dict)
		if list:
			return ListOptions( self.process_list(copied_options_dict) )
		else:
			return BaseOptions( self.process(copied_options_dict) )
			
	
	def process(self, options):
		new_options = {}
		new_options['embed'] = options.get('embed', None)
		if new_options['embed']:
			new_options['embed'] = set(new_options['embed'])
		new_options['allow_embedding'] = options.get('allow_embedding', True)
		new_options['fields'] = options.get('fields', None)
		if new_options['fields']:
			new_options['fields'] = set(new_options['fields'])
		new_options['show_hidden'] = options.get('show_hidden', False)
		new_options['context'] = options.get('context', {})
		new_options['bypass_authorization'] = options.get('bypass_authorization', False)
		
		if new_options['bypass_authorization']:
			new_options['can_show_hidden'] = True
		else:
			new_options['can_show_hidden'] = self.can_show_hidden(new_options['context'])
		
		return new_options
			
			
	def can_show_hidden(self, context):
		if self.hidden_field_authorization:
			if self.hidden_field_authorization.uses('identity') and 'identity' not in context:
				return False
			elif not self.hidden_field_authorization(context):
				return False
		return True
		
		
	def process_list(self, options):
		new_options = self.process(options)
		new_options['filter'] = options.get('filter', None)
		new_options['sort'] = options.get('sort', None)
		new_options['sort'] = options['sort'] if options.get('sort') else self.default_sort
		new_options['offset'] = options.get('offset', 0)
		new_options['limit'] = options.get('limit', 0)
		new_options['limit'] = new_options['limit'] if new_options['limit'] else self.default_limit
		if not new_options['bypass_authorization']:
			new_options['limit'] = min(new_options['limit'], self.max_limit)
		new_options['count'] = options.get('count', False)
		
		self.check_filter(new_options)
		self.check_sort(new_options)
		
		return new_options
		
		
	def check_filter(self, options):
		if not options['filter'] or options['bypass_authorization']:
			return
		if not self.enabled_filters:
			raise errors.CompoundValidationError({'filter':'Filtering is disabled.'})
		if options['can_show_hidden']:
			allowed_fields = self.enabled_filters
		else:
			allowed_fields = self.enabled_filters_no_hidden
		self.storage.check_filter(options['filter'], allowed_fields, options['context'])
		
		
	def check_sort(self, options):
		if not options['sort'] or options['bypass_authorization']:
			return
		if not self.enabled_sort:
			raise errors.CompoundValidationError({'sort':'Sorting is disabled.'})
		if options['can_show_hidden']:
			allowed_fields = self.enabled_sort
		else:
			allowed_fields = self.enabled_sort_no_hidden
			
		for k in options['sort']:
			field_name = k[1:]
			if field_name not in allowed_fields:
				raise errors.DisabledFieldError('The "%s" field cannot be used for sorting.' % field_name)


class BaseOptions(object):
	
	def __init__(self, options):
		self._options = options
		self._embed_by_class = {}
		
		
	def __getitem__(self, key):
		return self._options[key]
		
		
	def __getattr__(self, key):
		return self.__getitem__(key)
		
		
	def get_embed_for_type(self, base_entity, type):
		type = type.split('.')[-1]
		
		if type in self._embed_by_class:
			return self._embed_by_class[type]
		
		entity = None
		
		if type == base_entity.__name__:
			entity = base_entity
		else:
			for c in base_entity.children:
				if c.__name__ == type:
					entity = c
					break
		
		if entity is None:
			raise Exception, "Can't find the entity for '%s'" % type
		
		if self._options['embed']:
			embed = self._options['embed'].intersection(entity.embeddable)
		else:
			embed = entity.embed_by_default
		
		if self._options['fields']:
			embed.update( entity.embeddable.intersection( self._options['fields'] ) )
		
		if not self._options['show_hidden'] or not self._options['can_show_hidden']:
			embed.difference_update(entity.hidden_fields)
		
		self._embed_by_class[type] = (entity, embed)
		return entity, embed
		
		
		
class ListOptions(BaseOptions):
	pass



class InterfaceType(type):
	
	def __init__(cls, name, bases, attrs):
		if name == 'Interface':
			return
		
		members = dict(inspect.getmembers(cls))
		singular_name = members.get('singular_name')
		plural_name = members.get('plural_name')
		entity = members.get('entity')
		if entity:
			entity_name = entity.__name__.lower()
		else:
			raise Exception, "An interface must specify an entity to manage"
		
		if not singular_name and not plural_name:
			singular_name = entity_name
			plural_name = entity_name + 's'
		
		elif not singular_name:
			if plural_name[-1] == 's':
				singular_name = plural_name[:-1]
			else:
				singular_name = plural_name
		
		elif not plural_name:
			plural_name = singular_name + 's'
		
		storage = None
		for b in bases:
			if b.__name__ == 'Interface':
				storage = b.api.model.storage
				break
		
		cls.singular_name = singular_name
		cls.plural_name = plural_name
		cls.hooks = EventManager('create', 'update', 'delete')
		cls.rules = RuleSet(members.get('method_authorization'))
		cls.storage = storage
		
		hidden_fields = set(entity.hidden_fields.copy())
		for c in entity.children:
			hidden_fields.update(c.hidden_fields)
		
		cls.options_factory = OptionsFactory(
			storage=storage,
		    hidden_fields=hidden_fields, 
		    hidden_field_authorization=members.get('hidden_field_authorization'),
		    enabled_filters=members.get('enabled_filters', ()),
		    enabled_sort=members.get('enabled_sort', ()),
		    default_sort=members.get('default_sort', ()),
		    default_limit=members.get('default_limit', 0),
		    max_limit=members.get('max_limit', 100)
		)
		
		cls.api.add_interface(cls)
		
		
		
class Interface(object):
	
	__metaclass__ = InterfaceType
	
	
	# The entity class that will be exposed through this interface.
	entity = None
	
	# A dict of authorization rules. The key is a `cellardoor.method` or a tuple of methods 
	# and the value is an authorization rule.
	method_authorization = None
	
	# A `cellardoor.authorization.AuthenticationExpression` that must be met for hidden fields to be shown.
	hidden_field_authorization = None
	
	# A list or tuple of fields that are allowed to be used in filters.
	enabled_filters = ()
	
	# A list or tuple of fields that can be used in sorting.
	enabled_sort = ()
	
	# A list or tuple of fields that is a subset of `enabled_sort` that will be
	# used to sort results when no sort is supplied in the request.
	default_sort = ()
	
	default_limit = 0
	max_limit = 100
	
	
	def __init__(self):
		for method in ALL:
			if method not in self.rules.enabled_methods:
				setattr(self, method, self.disabled_method_error)
			
			
	def set_storage(self, storage):
		self.storage = storage
		self.options_factory.storage = storage
			
			
	def list(self, **kwargs):
		options = self.options_factory.create(kwargs, list=True)
		
		if not options['bypass_authorization']:
			self.rules.enforce_non_item_rules(LIST, options['context'])
		
		result = self.storage.get(self.entity, 
							filter=options.filter, sort=options.sort, 
							offset=options.offset, limit=options.limit,
							count=options.count)
		
		if options.count:
			return result
		
		if not options.bypass_authorization:
			self.rules.enforce_item_rules(LIST, result, options.context)
				
		return self.post(LIST, options, result)
		
		
	def create(self, fields, **kwargs):
		options = self.options_factory.create(kwargs)
		
		if not options.bypass_authorization:
			self.rules.enforce_non_item_rules(CREATE, options.context)
		
		self.entity.hooks.fire_before_create(fields, options.context)
		self.hooks.fire_before_create(fields, options.context)
		
		item = self.entity.validator.validate(fields)
		item['_id'] = self.storage.create(self.entity, item)
		
		if not options.bypass_authorization:
			self.rules.enforce_item_rules(CREATE, item, options.context)
		
		item = self.post(CREATE, options, item)
		
		self.entity.hooks.fire_after_create(item, options.context)
		self.hooks.fire_after_create(item, options.context)
		
		return item
		
		
	def get(self, id, **kwargs):
		options = self.options_factory.create(kwargs)
		
		if not options.bypass_authorization:
			self.rules.enforce_non_item_rules(GET, options.context)
		
		item = self.storage.get_by_id(self.entity, id)
		if item is None:
			raise errors.NotFoundError("No %s with id '%s' was found" % (self.singular_name, id))
		
		if not options.bypass_authorization:
			self.rules.enforce_item_rules(GET, item, options.context)
		
		return self.post(GET, options, item)
		
		
	def update(self, id, fields, _replace=False, _method=UPDATE, **kwargs):
		options = self.options_factory.create(kwargs)
		
		if not options.bypass_authorization:
			self.rules.enforce_non_item_rules(_method, options.context)
		
		self.entity.hooks.fire_before_update(id, fields, options.context)
		self.hooks.fire_before_update(id, fields, options.context)
		
		if not options.bypass_authorization and UPDATE in self.rules.item_rules:
			item = self.storage.get_by_id(self.entity, id)
			if item is None:
				raise errors.NotFoundError("No %s with id '%s' was found" % (self.singular_name, id))
			self.rules.enforce_item_rules(_method, item, options.context)
		
		fields = self.entity.validator.validate(fields, enforce_required=_replace)
		item = self.storage.update(self.entity, id, fields, replace=_replace)
		if item is None:
			raise errors.NotFoundError("No %s with id '%s' was found" % (self.singular_name, id))
		
		item = self.post(_method, options, item)
		
		self.entity.hooks.fire_after_update(item, options.context)
		self.hooks.fire_after_update(item, options.context)
		
		return item
		
		
	def replace(self, id, fields, **kwargs):
		return self.update(id, fields, _replace=True, _method=REPLACE, **kwargs)
		
		
	def delete(self, id, **kwargs):
		options = self.options_factory.create(kwargs)
		
		if not options.bypass_authorization:
			self.rules.enforce_non_item_rules(DELETE, options.context)
		
		item = self.storage.get_by_id(self.entity, id)
		if item is None:
			raise errors.NotFoundError("No %s with id '%s' was found" % (self.singular_name, id))
		
		if not options.bypass_authorization:
			self.rules.enforce_item_rules(DELETE, item, options.context)
		
		self.entity.hooks.fire_before_delete(id, options.context)
		self.hooks.fire_before_delete(id, options.context)
		
		self.storage.delete(self.entity, id)
		self.post(DELETE, options)
		
		options.context['item'] = item
		self.entity.hooks.fire_after_delete(id, options.context)
		self.hooks.fire_after_delete(id, options.context)
		
		
	def link(self, id, link_name, **kwargs):
		options = self.options_factory.create(kwargs)
		
		if not options.bypass_authorization:
			self.rules.enforce_non_item_rules(GET, options.context)
		
		item = self.storage.get_by_id(self.entity, id)
		if item is None:
			raise errors.NotFoundError("No %s with id '%s' was found" % (self.singular_name, id))
		
		if not options.bypass_authorization:
			self.rules.enforce_item_rules(GET, item, options.context)
			
		target_interface = self.get_linked_interface(link_name)
		if target_interface is None:
			raise errors.NotFoundError("The %s interface has no link '%s' defined" % (self.plural_name, link_name))
		link_field = getattr(self.entity, link_name)
		
		return target_interface.resolve_link(item, link_name, link_field, kwargs)
		
		
	def resolve_link(self, source_item, link_name, link_field, options):
		"""Get the item(s) pointed to by a link or link"""
		options = self.options_factory.create(options, list=True)
		
		if isinstance(link_field, InverseLink):
			return self._resolve_inverse_link(source_item, link_field, options)
		else:
			return self._resolve_link(source_item, link_name, link_field, options)
		
		
	def _resolve_inverse_link(self, source_item, link_field, options):
		"""
		Get the items for a single or multiple link
		"""
		options.filter = options.filter if options.filter else {}
		options.filter[link_field.field] = source_item['_id']
		
		if link_field.multiple:
			self.rules.enforce_non_item_rules(LIST, options.context)
			result = self.storage.get(self.entity, 
							filter=options.filter, sort=options.sort, 
							offset=options.offset, limit=options.limit,
							count=options.count)
			if options.count:
				return result
			self.rules.enforce_item_rules(LIST, result, options.context)
			return self.post(LIST, options, result)
		else:
			try:
				if not options.bypass_authorization:
					self.rules.enforce_non_item_rules(GET, options.context)
				item = next(iter(self.storage.get(self.entity, filter=options.filter, limit=1)))
				if not options.bypass_authorization:
					self.rules.enforce_item_rules(GET, item, options.context)
				return self.post(GET, options, item)
			except StopIteration:
				pass
		
	
	def _resolve_link(self, source_item, link_name, link_field, options):
		"""Get the items for a single or multiple link"""
		link_value = source_item.get(link_name)
		if link_value is None:
			return None
		
		if isinstance(link_field, ListOf):
			if not options.bypass_authorization:
				self.rules.enforce_non_item_rules(LIST, options.context)
			result = self.storage.get_by_ids(self.entity, link_value,
								filter=options.filter, sort=options.sort, 
								offset=options.offset, limit=options.limit,
								count=options.count)
			if options.count:
				return result
			if not options.bypass_authorization:
				self.rules.enforce_item_rules(LIST, result, options.context)
			
			if not options.sort and len(result) > 1:
				results_by_id = {}
				for r in result:
					results_by_id[r['_id']] = r
				new_result = []
				for id in link_value:
					item = results_by_id.get(id)
					if item:
						new_result.append(item)
				result = new_result
			return self.post(LIST, options, result)
		else:
			self.rules.enforce_non_item_rules(GET, options.context)
			item = self.storage.get_by_id(self.entity, link_value)
			self.rules.enforce_item_rules(GET, item, options.context)
			return self.post(GET, options, item)
			
			
	def get_linked_interface(self, link_name):
		link = None
		entities = [self.entity] + self.entity.children
		for entity in entities:
			link = entity.get_link(link_name)
			if link:
				break
		if not link:
			raise Exception, "Entity '%s' nor its children have a link called '%s'" % (self.entity.__name__, link_name)
			
		if isinstance(link, ListOf):
			linked_entity = link.field.entity
		else:
			linked_entity = link.entity
		return self.api.get_interface_for_entity(linked_entity)
		
		
	def prepare_item(self, item, options):
		self.remove_hidden_fields(item, options)
		self.add_embedded_links(item, options)
		return item
		
		
	def remove_hidden_fields(self, item, options):
		if not options.show_hidden or not options.can_show_hidden:
			for k in self.entity.hidden_fields:
				item.pop(k, None)
		if options.fields is not None:
			for k in item.keys():
				if k not in options.fields and not k.startswith('_'):
					item.pop(k, None)
			
		
	def add_embedded_links(self, item, options):
		"""Add embedded links and links to an item"""
		if not options.allow_embedding:
			return
		
		entity, embed = options.get_embed_for_type(self.entity, item.get('_type', self.entity.__name__))
		
		for link_name in embed:
			linked_interface = self.get_linked_interface(link_name)
			if not linked_interface:
				raise Exception, "No link defined in '%s' interface for embedded link '%s'" % (self.plural_name, link_name)
			
			link_field = getattr(entity, link_name)
			
			link_options = {
				'context': options.context,
				'allow_embedding': False,
				'show_hidden': options.show_hidden
			}
			
			embedded_fields = link_field.field.embedded_fields if isinstance(link_field, ListOf) else link_field.embedded_fields
			if embedded_fields:
				link_options['fields'] = embedded_fields
			
			result = linked_interface.resolve_link(item, link_name, link_field, link_options)
			
			if result:
				item[link_name] = result
				
		return item
		
		
	def post(self, method, options, result=None):
		"""Perform post-method hooks including authentication that requires fetched items."""
		if result is None:
			return
			
		if method == LIST:
			new_results = []
			for item in result:
				new_results.append(
					self.prepare_item(item, options)
				)
			options.context['item'] = result
			return new_results
		else:
			options.context['item'] = result
			return self.prepare_item(result, options)
		
		
	def disabled_method_error(self, *args, **kwargs):
		raise errors.DisabledMethodError, "This method is not enabled."
		
		


from functools import partial


class AuthorizationExpression(object):
	
	def __call__(self, context):
		raise NotImplementedError
		
		
	def __and__(self, other):
		self.require_auth_expr(other)
		return AndExpression(self, other)
		
		
	def __or__(self, other):
		self.require_auth_expr(other)
		return OrExpression(self, other)
		
		
	def __invert__(self):
		return lambda x: not self(x)
		
		
	def __eq__(self, other):
		raise NotImplementedError
		
		
	def require_auth_expr(self, other):
		if not isinstance(other, AuthorizationExpression):
			raise TypeError, "Type '%s' is incompatible with type '%s'" % (type(self).__name__, type(other).__name__)
			
			
	def uses(self, key):
		raise NotImplementedError
			

			
class BinaryExpression(AuthorizationExpression):
	
	def __init__(self, a, b):
		self.a = a
		self.b = b
	
	def uses(self, key):
		return self.a.uses(key) or self.b.uses(key)
		
		
	def __eq__(self, other):
		return isinstance(other, self.__class__) and other.a == self.a and other.b == self.b
		
		
			
class AndExpression(BinaryExpression):
	
	
	def __call__(self, context):
		return self.a(context) and self.b(context)
		
		
	def __repr__(self):
		return 'AndExpression(%s, %s)' % (repr(self.a), repr(self.b))
		
		
		
class OrExpression(BinaryExpression):
		
	def __call__(self, context):
		return self.a(context) or self.b(context)
		
		
	def __repr__(self):
		return 'OrExpression(%s, %s)' % (repr(self.a), repr(self.b))
		


class ObjectProxy(AuthorizationExpression):
	
	def __init__(self, name):
		self._name = name
		
		
	def __eq__(self, other):
		return isinstance(other, ObjectProxy) and other._name == self._name
		
		
	def __getattr__(self, key):
		return ObjectProxyValue(self, key)
		
		
	def __getitem__(self, key):
		return self.__getattr__(key)
		
		
	def __call__(self, context):
		return self._name in context
		
		
	def __repr__(self):
		return 'ObjectProxy(%s)' % self._name
		
		
	def match(self, fn):
		return ObjectProxyMatch(self, fn)
		
		
	def exists(self):
		return self
		
		
	def uses(self, key):
		return self._name == key
		
		
	def get(self, context):
		return context.get(self._name, {})
		
		
		
class ObjectProxyMatch(AuthorizationExpression):
	
	def __init__(self, proxy, fn):
		self._proxy = proxy
		self.fn = fn
		
		
	def __repr__(self):
		return 'ObjectProxyMatch(%s, %s)' % (repr(self._proxy), repr(self.fn))
		
		
	def __eq__(self, other):
		return isinstance(other, ObjectProxyMatch) and other._proxy == self._proxy and other.fn == self.fn
		
		
	def __call__(self, context):
		obj = self._proxy.get(context)
		return self.fn(obj)
		
		
	def uses(self, key):
		return self._proxy.uses(key)
		
		
		
class ObjectProxyValue(AuthorizationExpression):
	
	def __init__(self, proxy, key):
		self._proxy = proxy
		self._key = key
		
		
	def __repr__(self):
		return 'ObjectProxyValue(%s, %s)' % (repr(self._proxy), self._key)
		
		
	def __eq__(self, other):
		return isinstance(other, ObjectProxyValue) and other._proxy == self._proxy and other._key == self._key
		
		
	def get_value(self, context):
		obj = self._proxy.get(context)
		val = obj.get(self._key)
		return val
		
		
	def exists(self):
		return self
		
		
	def uses(self, key):
		return self._proxy.uses(key)
		
		
	def __call__(self, context):
		return self._proxy(context) and self._key in self._proxy.get(context)
		
		
	def __eq__(self, other):
		return EqualsComparison(self, other)
		
		
	def __ne__(self, other):
		return NotEqualsComparison(self, other)
		
		
	def __lt__(self, other):
		return LessThanComparison(self, other)
		
		
	def __gt__(self, other):
		return GreaterThanComparison(self, other)
		
		
	def __le__(self, other):
		return LessThanEqualComparison(self, other)
		
		
	def __ge__(self, other):
		return GreaterThanEqualComparison(self, other)
		
		
	def in_(self, other):
		return ContainsComparison(self, other)
		
		
class ObjectProxyValueComparison(AuthorizationExpression):
	
	opstr = ''
	
	def __init__(self, proxy, other):
		self._proxy = proxy
		self.other = other
		
		
	def __repr__(self):
		return '%s %s %s' % (repr(self._proxy), self.opstr, repr(self.other))
		
	
	def __eq__(self, other):
		return isinstance(other, self.__class__) and other._proxy == self._proxy and other.other == self.other
		
		
	def __call__(self, context):
		a = self._proxy.get_value(context)
		
		if isinstance(self.other, ObjectProxyValue):
			b = self.other.get_value(context)
		else:
			b = self.other
			
		return self.compare(a, b)
			
		
	def compare(self, a, b):
		raise NotImplementedError
		
		
	def uses(self, key):
		if isinstance(self.other, AuthorizationExpression):
			return self._proxy.uses(key) or self.other.uses(key)
		else:
			return self._proxy.uses(key)
		
		
		
class EqualsComparison(ObjectProxyValueComparison):
	opstr = '=='
	
	def compare(self, a, b):
		return a == b
		
		
class NotEqualsComparison(ObjectProxyValueComparison):
	opstr = '!='
	
	def compare(self, a, b):
		return a != b
		
		
class LessThanComparison(ObjectProxyValueComparison):
	opstr = '<'
	
	def compare(self, a, b):
		return a < b
		
		
class GreaterThanComparison(ObjectProxyValueComparison):
	opstr = '>'
	
	def compare(self, a, b):
		return a > b
		
		
class LessThanEqualComparison(ObjectProxyValueComparison):
	opstr = '<='
	
	def compare(self, a, b):
		return a <= b
		
		
class GreaterThanEqualComparison(ObjectProxyValueComparison):
	opstr = '>='
	
	def compare(self, a, b):
		return a >= b
		
		
class ContainsComparison(ObjectProxyValueComparison):
	opstr = ' in '
	
	def compare(self, a, b):
		return a in b
		
		
		
class ItemProxy(ObjectProxy):
	
	def __init__(self, entity, name='item'):
		super(ItemProxy, self).__init__(name)
		self._entity = entity
		
		
	def __getattr__(self, key):
		link = self._entity.get_link(key)
		if link:
			return LinkProxy(self, link.entity, key)
		else:
			return ObjectProxyValue(self, key)
		
		
	def __repr__(self):
		return 'ItemProxy(%s, %s)' % (self._entity.__name__, self._name)
		
	
	
class LinkProxy(ObjectProxyValue, ItemProxy):
	
	def __init__(self, proxy, entity, name):
		ItemProxy.__init__(self, entity, name)
		ObjectProxyValue.__init__(self, proxy, name)
		
		
	def get(self, context):
		item = self._proxy.get(context)
		interface = context['api'].get_interface_for_entity(self._proxy._entity)
		return interface.link(item['_id'], self._name, 
							bypass_authorization=True, show_hidden=True)
		
		
	def __repr__(self):
		return 'LinkProxy(%s, %s)' % (self._proxy._entity.__name__, self._name)
		


from .. import model


class EntitySerializer(object):
	
	fallbacks = (
		model.Text, model.Integer, model.Float, model.DateTime,
		model.Boolean, model.Enum, model.ListOf, model.OneOf, 
		model.Compound, model.Anything
	)
		
	def create_schema(self, entity):
		props = {}
		required_props = []
		for k,v in entity.fields.items():
			props[k] = self.get_property(v)
			if v.required:
				required_props.append(k)
		definition =  {
			'title': entity.__name__,
			'properties': props,
			'links': self.get_links(entity)
		}
		if required_props:
			definition['required'] = required_props
		return definition
			
			
	def get_links(self, entity):
		if len(entity.hierarchy) > 0:
			parent = entity.hierarchy[-1]
			return {
				'parent': {
					'rel': 'parent',
					'href': '#/definitions/%s' % parent.__name__
				}
			}
		else:
			return {}
			
			
	def get_property(self, field):
		prop = {}
		prop['default'] = field.default
		prop['format'] = field.__class__.__name__
		if field.label:
			prop['title'] = field.label
		if field.description:
			prop['description'] = field.description
		type_name = field.__class__.__name__
		method_name = 'handle_%s' % type_name
		if not hasattr(self, method_name):
			method_name = None
			for cls in self.fallbacks:
				if isinstance(field, cls):
					method_name = 'handle_%s' % cls.__name__
					break
			if not method_name:
				raise ValueError, "No handler for %s" % type_name
		getattr(self, method_name)(field, prop)
		return prop
		
		
	def handle_Text(self, field, prop):
		prop['type'] = 'string'
		if field.maxlength:
			prop['maxLength'] = field.maxlength
		if field.minlength:
			prop['minLength'] = field.minlength
		if field.regex:
			prop['pattern'] = field.regex.pattern
		
		
	def handle_Integer(self, field, prop):
		prop['type'] = 'integer'
		if field.min:
			prop['minimum'] = field.min
		if field.max:
			prop['maximum'] = field.max
		
		
	def handle_Float(self, field, prop):
		self.handle_Integer(field, prop)
		prop['type'] = 'number'
		
		
	def handle_DateTime(self, field, prop):
		prop['type'] = 'string'
		
		
	def handle_Email(self, field, prop):
		self.handle_Text(field, prop)
		del prop['pattern']
		
		
	def handle_URL(self, field, prop):
		self.handle_Text(field, prop)
		del prop['pattern']
		
		
	def handle_Boolean(self, field, prop):
		prop['type'] = 'boolean'
		
		
	def handle_Enum(self, field, prop):
		prop['enum'] = field.values
		
		
	def handle_ListOf(self, field, prop):
		prop['type'] = 'array'
		prop['items'] = self.get_property(field.field)
		
		
	def handle_OneOf(self, field, prop):
		prop['anyOf'] = map(self.get_property, field.fields)
		
		
	def handle_Link(self, field, prop):
		self.handle_Text(field, prop)
		prop['format'] = 'Link'
		prop['schema'] = '#/definitions/%s' % field.entity.__name__
		
		
	def handle_Compound(self, field, prop):
		prop['type'] = 'object'
		properties = {}
		required = []
		for k, v in field.fields.items():
			properties[k] = self.get_property(v)
			if v.required:
				required.append(k)
		prop['properties'] = properties
		if len(required) > 0:
			prop['required'] = required
			
			
	def handle_Anything(self, field, prop):
		prop['anyOf'] = [
			{'type':'array'},
			{'type':'boolean'},
			{'type':'null'},
			{'type':'object'},
			{'type':'string'},
			{'type':'number'}
		]
			
			
	def handle_BoundingBox(self, field, prop):
		prop['type'] = 'array'
		prop['items'] = {
			'type': 'float',
			'minimum': -180.0,
			'maximum': 180.0,
			'maxItems': 4,
			'minItems': 4
		}
		
		
	def handle_LatLng(self, field, prop):
		prop['type'] = 'array'
		prop['items'] = {
			'type': 'float',
			'minimum': -180.0,
			'maximum': 180.0,
			'maxItems': 2,
			'minItems': 2
		}
		
		
	def handle_TypeOf(self, field, prop):
		type = field.types[0]
		if type == dict:
			prop['type'] = 'object'
		elif type == list:
			prop['type'] = 'array'
		elif len(field.types) == 2 and int in field.types and float in field.types:
			prop['type'] = 'number'
		elif type == int:
			prop['type'] = 'integer'
		elif type == float:
			prop['type'] = 'float'
		elif type == unicode:
			prop['type'] = 'string'
		
		
		
class APISerializer(object):
	
	def create_schema(self, api, base_url, entity_serializer):
		self.base_url = base_url
		definitions = {}
		for name, entity in api.model.entities.items():
			definitions[name] = entity_serializer.create_schema(entity)
		
		resources = {}
		for name, interface in api.interfaces.items():
			entity_links = definitions[interface.entity.__name__]['links']
			if 'resource' not in entity_links:
				entity_links['resource'] = {
					'rel': 'resource',
					'href': '#/properties/%s' % interface.plural_name
				}
			resources[interface.plural_name] = self.get_resource_schema(interface)
		
		return {
			"$schema": "http://json-schema.org/draft-04/schema#",
			"type": "object",
			"definitions": definitions,
			"properties": resources
		}
		
		
	def get_resource_schema(self, interface):
		links = {}
		
		for method in interface.rules.enabled_methods:
			fn = getattr(self, 'get_%s_link' % method)
			links[method] = fn(interface)
			
		for k,v in interface.entity.links.items():
			links['link-%s' % k] = self.get_link_link(interface, k, v)
		
		return {
			"title": interface.plural_name,
			"links": links
		}
		
		
	def get_list_link(self, interface):
		return {
			'href': self.base_url + '/%s' % interface.plural_name,
			'method': 'GET',
			'rel': 'instances',
			'title': 'List',
			'targetSchema': {
				'type': 'array',
				'items': {
					'$ref': self.entity_schema_ref(interface)
				}
			}
		}
		
		
	def get_get_link(self, interface):
		return {
			'href': self.base_url + '/%s/{id}' % interface.plural_name,
			'method': 'GET',
			'rel': 'instance',
			'title': 'Details',
			'targetSchema': { '$ref': self.entity_schema_ref(interface) }
		}
		
		
	def get_create_link(self, interface):
		return {
			'href': self.base_url + '/%s' % interface.plural_name,
			'method': 'POST',
			'rel': 'create',
			'title': 'New',
			'schema': { '$ref': self.entity_schema_ref(interface) },
			'targetSchema': { '$ref': self.entity_schema_ref(interface) }
		}
		
		
	def get_update_link(self, interface):
		return {
			'href': self.base_url + '/%s/{id}' % interface.plural_name,
			'method': 'PATCH',
			'rel': 'update',
			'title': 'Update',
			'schema': {
				'allOf': [ { '$ref': self.entity_schema_ref(interface) } ],
				'required': []
			},
			'targetSchema': { '$ref': self.entity_schema_ref(interface) }
		}
		
		
	def get_replace_link(self, interface):
		return {
			'href': self.base_url + '/%s/{id}' % interface.plural_name,
			'method': 'PUT',
			'rel': 'replace',
			'title': 'Replace',
			'schema': { '$ref': self.entity_schema_ref(interface) },
			'targetSchema': { '$ref': self.entity_schema_ref(interface) }
		}
		
		
	def get_delete_link(self, interface):
		return {
			'href': self.base_url + '/%s/{id}' % interface.plural_name,
			'method': 'DELETE',
			'rel': 'delete',
			'title': 'Delete'
		}
		
		
	def get_link_link(self, interface, link_name, link_interface):
		schema_link = {
			'href': self.base_url + '/%s/{id}/%s' % (interface.plural_name, link_name),
			'method': 'GET',
			'rel': 'link',
			'title': 'Link'
		}
		
		link = None
		entities = [interface.entity] + interface.entity.children
		for entity in entities:
			link = entity.fields.get(link_name)
			if not link:
				link = entity.links.get(link_name)
			if link:
				break
		
		if interface.entity.is_multiple_link(link):
			schema_link['targetSchema'] = {
				'type': 'array',
				'items': { '$ref': self.entity_schema_ref(link_interface) }
			}
		else:
			schema_link['targetSchema'] = { '$ref': self.entity_schema_ref(link_interface) }
			
		return schema_link
		
		
	def entity_schema_ref(self, interface):
		return '#/definitions/%s' % interface.entity.__name__


def to_jsonschema(api, base_url, api_cls=APISerializer, entity_cls=EntitySerializer):
	api_serializer = api_cls()
	entity_serializer = entity_cls()
	return api_serializer.create_schema(api, base_url, entity_serializer)
	
	
	


import functools
import falcon
import logging
import inspect
from ..api.methods import LIST, CREATE, GET, REPLACE, UPDATE, DELETE, get_http_methods
from ..serializers import JSONSerializer, MsgPackSerializer
from ..views import View
from ..views.minimal import MinimalView
from .. import errors

__all__ = ['add_to_falcon']


class Resource(object):
	"""
	A resource exposes an interface through falcon
	"""
	
	# The types of content that are accepted.
	accept_serializers = (JSONSerializer(), MsgPackSerializer())
	params_serializer = JSONSerializer()
	
	
	def __init__(self, interface, views):
		self.interface = interface
		self.views = views
		self.logger = logging.getLogger(__name__)
		
		
	def add_to_falcon(self, app):
		methods = self.interface.rules.enabled_methods
		interface_methods = methods.intersection((LIST, CREATE))
		individual_methods = methods.intersection((GET, REPLACE, UPDATE, DELETE))
		
		if not interface_methods and not individual_methods:
			raise Exception, "The '%s' interface exposes no methods" % self.interface.plural_name
		
		if interface_methods:
			app.add_route('/%s' % self.interface.plural_name, ListEndpoint(self, interface_methods))
			
		if individual_methods:
			app.add_route('/%s/{id}' % self.interface.plural_name, IndividualEndpoint(self, individual_methods))
		
		for link_name, link in self.interface.entity.get_links().items():
			if self.interface.api.get_interface_for_entity(link.entity):
				app.add_route('/%s/{id}/%s' % (self.interface.plural_name, link_name), 
					ReferenceEndpoint(self, link_name))
			
	
	def list(self, req, resp):
		kwargs = self.get_kwargs(req)
		items = self.interface.list(**kwargs)
		self.send_list(req, resp, items)
		
		
	def count(self, req, resp):
		kwargs = self.get_kwargs(req)
		kwargs['count'] = True
		result = self.interface.list(**kwargs)
		resp.content_type, _ = View.choose(req, self.views)
		resp.set_header('X-Count', str(result))
		
		
	def create(self, req, resp):
		fields = self.get_fields_from_request(req)
		kwargs = self.get_kwargs(req, 'show_hidden', 'context', 'embedded')
		item = self.interface.create(fields, **kwargs)
		resp.status = falcon.HTTP_201
		self.send_one(req, resp, item)
		
	
	def get(self, req, resp, id):
		kwargs = self.get_kwargs(req, 'show_hidden', 'context', 'embedded')
		item = self.interface.get(id, **kwargs)
		self.send_one(req, resp, item)
		
		
	def update(self, req, resp, id):
		fields = self.get_fields_from_request(req)
		kwargs = self.get_kwargs(req, 'show_hidden', 'context', 'embedded')
		item = self.interface.update(id, fields, **kwargs)
		self.send_one(req, resp, item)
		
		
	def replace(self, req, resp, id):
		fields = self.get_fields_from_request(req)
		kwargs = self.get_kwargs(req, 'show_hidden', 'context', 'embedded')
		item = self.interface.replace(id, fields, **kwargs)
		self.send_one(req, resp, item)
		
		
	def delete(self, req, resp, id):
		self.interface.delete(id, context=self.get_context(req))
		
		
	def get_link_or_reference(self, req, resp, id, link_name):
		kwargs = self.get_kwargs(req)
		result = self.interface.link(id, link_name, **kwargs)
		if isinstance(result, dict):
			self.send_one(req, resp, result)
		else:
			self.send_list(req, resp, result)
			
			
	def count_link_or_reference(self, req, resp, id, link_name):
		kwargs = self.get_kwargs(req)
		kwargs['count'] = True
		result = self.interface.link(id, link_name, **kwargs)
		resp.content_type, _ = View.choose(req, self.views)
		if isinstance(result, int):
			resp.set_header('X-Count', str(result))
		
		
	def send_one(self, req, resp, item):
		resp.content_type, resp.body = self.serialize_one(req, item)
		
		
	def send_list(self, req, resp, items):
		resp.content_type, resp.body = self.serialize_list(req, items)
		
		
	def serialize_one(self, req, data):
		return self.serialize(req, 'get_individual_response', data)
		
			
	def serialize_list(self, req, data):
		return self.serialize(req, 'get_list_response', data)
		
		
	def serialize(self, req, method_name, data):
		view = self.get_view(req)
		method = getattr(view, method_name)
		try:
			return method(req, data)
		except Exception, e:
			self.logger.exception('Failed to serialize response.')
			raise falcon.HTTPInternalServerError('Internal Server Error', '')
			
			
	def get_view(self, req):
		"""Get the correct view based on the Accept header"""
		_, view = View.choose(req, self.views)
		return view
			
			
	def get_fields_from_request(self, req):
		"""Unserializes the request body based on the request's content type"""
		for serializer in self.accept_serializers:
			if req.content_type.startswith( serializer.mimetype ):
				try:
					return serializer.unserialize(req.stream)
				except Exception:
					self.logger.exception('Could not parse request body.')
					raise falcon.HTTPBadRequest('Bad Request', 'Could not parse request body.')
		raise falcon.HTTPUnsupportedMediaType(
			'The supported types are: %s' % ', '.join([x.mimetype for x in self.accept_serializers]))
		
		
	def get_kwargs(self, req, *include):
		"""Parse out the filter, sort, etc., parameters from a request"""
		params = (
			('embedded', self.params_serializer.unserialize_string, None),
			('filter', self.params_serializer.unserialize_string, None),
			('sort', self.params_serializer.unserialize_string, None),
			('offset', int, 0),
			('limit', int, 0),
			('show_hidden', self.bool_field, False)
		)
		results = {}
		if len(include) > 0:
			include = set(include)
		else:
			include = None
		for name, fn, default in params:
			if include and name not in include:
				continue
			results[name] = self.get_param(req, name, fn, default=default)
			
		if not include or 'context' in include:
			results['context'] = self.get_context(req)
		return results
		
		
	def bool_field(self, value):
		return True if value.lower() == 'true' or value == '1' else False
		
		
	def get_param(self, req, param_name, parsing_fn, default=None):
		"""Get a parsed query param"""
		param = req.get_param(param_name)
		if not param:
			return default
		try:
			return parsing_fn(param)
		except Exception, e:
			raise falcon.HTTPBadRequest("Bad Request", "Could not parse %s parameter" % param_name)
			
			
	def get_context(self, req):
		context = {}
		identity = req.env.get('cellardoor.identity')
		if identity:
			context['identity'] = identity
		return context
		
			
			
			
class Endpoint(object):
	
	def __init__(self, resource, methods):
		self.resource = resource
		self.register_methods(methods)
		
		
	def register_methods(self, methods):
		for method in methods:
			fn = getattr(self, method)
			for http_method in get_http_methods(method):
				setattr(self, 'on_%s' % http_method, fn)
		if LIST in methods:
			setattr(self, 'on_head', self.count)
			

class ListEndpoint(Endpoint):
	
	def list(self, req, resp):
		return self.resource.list(req, resp)
		
		
	def count(self, req, resp):
		return self.resource.count(req, resp)
		
		
	def create(self, req, resp):
		return self.resource.create(req, resp)
		
		
class IndividualEndpoint(Endpoint):
	
		
	def get(self, req, resp, id):
		return self.resource.get(req, resp, id)
		
		
	def update(self, req, resp, id):
		return self.resource.update(req, resp, id)
		
		
	def replace(self, req, resp, id):
		return self.resource.replace(req, resp, id)
		
		
	def delete(self, req, resp, id):
		return self.resource.delete(req, resp, id)
		
		
class ReferenceEndpoint(object):
	
	def __init__(self, resource, link_name):
		self.resource = resource
		self.link_name = link_name
		
		
	def on_get(self, req, resp, id):
		return self.resource.get_link_or_reference(req, resp, id, self.link_name)
		
		
	def on_head(self, req, resp, id):
		return self.resource.count_link_or_reference(req, resp, id, self.link_name)
		
		
		
def not_found_handler(exc, req, resp, params):
	raise falcon.HTTPNotFound()


def not_authenticated_handler(exc, req, resp, params):
	raise falcon.HTTPUnauthorized('Unauthorized', 'You must authenticate to access this resource.')
	
	
def not_authorized_handler(exc, req, resp, params):
	raise falcon.HTTPForbidden('Forbidden', 'You are not allowed to access this resource.')
	
	
def validation_error_handler(views, exc, req, resp, params):
	_, view = View.choose(req, views)
	resp.content_type, resp.body = view.get_individual_response(req, exc.errors)
	resp.status = falcon.HTTP_400
	
	
def disabled_field_error(exc, req, resp, params):
	raise falcon.HTTPUnauthorized('Unauthorized', exc.message)
	
	
def duplicate_field_error(views, exc, req, resp, params):
	error = {}
	error[exc.message] = 'A duplicate value already exists.'
	_, view = View.choose(req, views)
	resp.content_type, resp.body = view.get_individual_response(req, error)
	resp.status = falcon.HTTP_400
		
		
class FalconApp(object):
	
	def __init__(self, api, falcon_app=None, views=(MinimalView,)):
		if falcon_app is None:
			falcon_app = falcon.API()
		self.falcon_app = falcon_app
		self.api = api
		self.resources = {}
		
		views_by_type = []
		
		for v in views:
			if inspect.isclass(v):
				v = v()
			for mimetype, _ in v.serializers:
				views_by_type.append((mimetype, v))
				
		falcon_app.add_error_handler(errors.NotFoundError, not_found_handler)
		falcon_app.add_error_handler(errors.NotAuthenticatedError, not_authenticated_handler)
		falcon_app.add_error_handler(errors.NotAuthorizedError, not_authorized_handler)
		validation_error_handler_with_views = functools.partial(validation_error_handler, views_by_type)
		falcon_app.add_error_handler(errors.CompoundValidationError, validation_error_handler_with_views)
		falcon_app.add_error_handler(errors.DisabledFieldError, disabled_field_error)
		duplicate_field_error_with_views = functools.partial(duplicate_field_error, views_by_type)
		falcon_app.add_error_handler(errors.DuplicateError, duplicate_field_error_with_views)
		
		for interface in api.interfaces.values():
			resource = Resource(interface, views_by_type)
			resource.add_to_falcon(falcon_app)
			self.resources[interface.plural_name] = resource
			
			
	def __call__(self, *args, **kwargs):
		return self.falcon_app(*args, **kwargs)


import re
import pymongo
from datetime import datetime
from bson.objectid import ObjectId
from . import Storage
from .. import errors

find_dupe_index_pattern = re.compile(r'\$([a-zA-Z0-9_]+)\s+')

class MongoDBStorage(Storage):
	
	special_fields = { '$where', '$text' }
	
	def __init__(self, db=None, *args, **kwargs):
		self.client = pymongo.MongoClient(*args, **kwargs)
		self.db = self.client[db]
		self.unique_fields_by_index = {}
		
		
	def setup(self, model):
		for e in model.entities.values():
			collection = self.get_collection(e)
			for k,v in e.fields.items():
				if v.unique:
					index_name = collection.ensure_index(k, unique=True, sparse=True)
					self.unique_fields_by_index[index_name] = k
		
	
	def get(self, entity, filter=None, fields=None, sort=None, offset=0, limit=0, count=False):
		to_dict = self.document_to_dict
		if filter and '_id' in filter and isinstance(filter['_id'], basestring):
			filter['_id'] = self._objectid(filter['_id'])
		
		sort_pairs = []
		if filter and '$text' in filter:
			sort_pairs.append(('score', {'$meta':'textScore'}))
			if not fields:
				fields = {}
			fields['score'] = {'$meta':'textScore'}
		if sort:
			sort_pairs.extend([(field[1:], 1) if field[0] == '+' else (field[1:], -1) for field in sort])
		
		collection = self.get_collection(entity)
		
		type_filter = self.get_type_filter(entity)
		if type_filter:
			if not filter:
				filter = type_filter
			else:
				filter.update(type_filter)
		
		results = collection.find(spec=filter, 
								  fields=fields, 
								  sort=sort_pairs, 
								  skip=offset, 
								  limit=limit)
		
		if count:
			return results.count()
		else:
			return map(to_dict, results)
			
			
	def get_by_ids(self, entity, ids, filter=None, fields=None, sort=None, offset=0, limit=0, count=False):
		if not filter:
			filter = {}
		filter['_id'] = {'$in':map(self._objectid, ids)}
		return self.get(entity, filter=filter, fields=fields, sort=sort, offset=offset, limit=limit, count=count)
		
		
	def get_by_id(self, entity, id, filter=None, fields=None):
		collection = self.get_collection(entity)
		filter = filter if filter else {}
		filter['_id'] = self._objectid(id)
		type_filter = self.get_type_filter(entity)
		if type_filter:
			filter.update(type_filter)
		result = collection.find_one(filter, fields=fields)
		
		if result is None:
			return None
		else:
			return self.document_to_dict(result)
		
		
	def create(self, entity, fields):
		collection = self.get_collection(entity)
		type_name = self.get_type_name(entity)
		if type_name:
			fields['_type'] = type_name
		if '_id' in fields:
			fields['_id'] = self._objectid(fields['_id'])
		try:
			obj_id = collection.insert(fields.copy())
		except pymongo.errors.DuplicateKeyError, e:
			self._raise_dupe_error(e)
			
		return self._from_objectid(obj_id)
		
		
	def update(self, entity, id, fields, replace=False):
		try:
			collection = self.get_collection(entity)
			obj_id = self._objectid(id)
			if replace:
				doc = fields
			else:
				doc = { '$set': fields }
			doc = collection.find_and_modify({ '_id': obj_id }, doc, new=True)
			if doc:
				return self.document_to_dict(doc)
		except pymongo.errors.DuplicateKeyError, e:
			self._raise_dupe_error(e)
			
			
	def delete(self, entity, id, deleted_by=None):
		collection = self.get_collection(entity)
		obj_id = self._objectid(id)
		collection.remove(obj_id)
		
		
	def document_to_dict(self, doc):
		doc['_id'] = self._from_objectid(doc['_id'])
		return doc
		
		
	def get_collection(self, entity):
		if len(entity.hierarchy) > 0:
			collection_name = entity.hierarchy[0].__name__
		else:
			collection_name = entity.__name__
			
		# We use getattr here instead of __getitem__ to
		# make it easier to inject mock collections objects
		# for testing
		return getattr(self.db, collection_name)
			
			
	def get_type_name(self, entity):
		if len(entity.hierarchy) > 0:
			return '.'.join([x.__name__ for x in entity.hierarchy]) + '.' + entity.__name__
		else:
			return None
			
			
	def get_type_filter(self, entity):
		type_name = self.get_type_name(entity)
		if type_name:
			return {'_type':{'$regex':'^%s' % re.escape(type_name)}}
			
		
	def check_filter(self, filter, allowed_fields, context):
		allowed_fields = set(allowed_fields)
		return self._check_filter(filter, allowed_fields, context)
		
		
	def _check_filter(self, filter, allowed_fields, context):
		if not isinstance(filter, dict):
			return
		for k,v in filter.items():
			
			if k.startswith('$'):
				if k in self.special_fields:
					continue
			elif k not in allowed_fields:
				raise errors.DisabledFieldError('You cannot filter by the "%s" field' % k)
			
			identity_value = self._get_identity_value(v, context)
			if identity_value:
				filter[k] = identity_value
			elif isinstance(v, (list, tuple)):
				new_v = []
				for x in v:
					identity_value = self._get_identity_value(x, context)
					if identity_value:
						new_v.append(identity_value)
					else:
						new_v.append(x)
					self._check_filter(x, allowed_fields, context)
					filter[k] = new_v
			elif isinstance(v, dict):
				self._check_filter(v, allowed_fields, context)
				
	def _get_identity_value(self, key, context):
		if isinstance(key, basestring):
			if key.startswith('$identity'):
				try:
					return reduce(dict.get, key[1:].split("."), context)
				except:
					raise errors.CompoundValidationError({'filter': 'Attempting to use a non-existent context variable: %s' % key})
				
				
	def _raise_dupe_error(self, orig_exc):
		m = find_dupe_index_pattern.search(orig_exc.message)
		if m:
			index_name = m.group(1)
			key_name = self.unique_fields_by_index.get(index_name)
		else:
			key_name = 'unknown'
		
		raise errors.DuplicateError(key_name)
			
		
	def _objectid(self, id):
		try:
			return ObjectId(id)
		except:
			return str(id)
			
	def _from_objectid(self, id):
		if isinstance(id, ObjectId):
			return str(id)
		else:
			return id


import base64
from . import Identifier
from .. import errors


class BasicAuthIdentifier(Identifier):
	"""
	Identifies Basic auth credentials in the Authorization header.
	
	The credentials take the form `{'username':'foo', 'password':'bar'}`
	"""
	type = 'Basic'
	
	def identify(self, environ):
		data = environ.get('HTTP_AUTHORIZATION')
		if not data:
			return None
		parts = data.split()
		if len(parts) != 2:
			return None
		if parts[0] != self.type:
			return None
		try:
			decoded_data = base64.standard_b64decode(parts[1])
		except TypeError:
			raise errors.IdentificationError
		else:
			try:
				username, password = decoded_data.split(':')
			except ValueError:
				raise errors.IdentificationError
			else:
				return {'username':username, 'password':password}


from .basic import BasicAuthIdentifier


class TokenAuthIdentifier(BasicAuthIdentifier):
	type = 'Token'
	
	def identify(self, environ):
		credentials = super(TokenAuthIdentifier, self).identify(environ)
		if credentials:
			return {
				'id': credentials['username'],
				'token': credentials['password']
			}


__all__ = (
	'LIST',
	'GET',
	'CREATE',
	'UPDATE',
	'REPLACE',
	'DELETE',
	'ALL',
	'get_http_methods'
)


LIST = 'list'
GET = 'get'
CREATE = 'create'
REPLACE = 'replace'
UPDATE = 'update'
DELETE = 'delete'

ALL = (LIST, GET, CREATE, UPDATE, REPLACE, DELETE)

_http_methods = {
	LIST: ('get',),
	GET: ('get',),
	CREATE: ('post',),
	REPLACE: ('put',),
	UPDATE: ('patch',),
	DELETE: ('delete',)
}

def get_http_methods(method):
	return _http_methods[method]
	


from datetime import datetime
from .fields import DateTime


__all__ = ['Timestamped']


class Timestamped(object):
    
    created = DateTime(label="Created")
    modified = DateTime(label="Modified")
    
    def before_create(self, fields, *args, **kwargs):
        now = datetime.utcnow()
        fields['created'] = now
        fields['modified'] = now
        
        
    def before_update(self, id, fields, *args, **kwargs):
        fields['modified'] = datetime.utcnow()
        